from PySide6 import QtWidgets, QtCore, QtGui

from ..constants import NodeEnum, ViewerNavEnum


class NodeNavigationDelagate(QtWidgets.QStyledItemDelegate):

    def paint(self, painter, option, index):
        """
        Args:
            painter (QtGui.QPainter):
            option (QtGui.QStyleOptionViewItem):
            index (QtCore.QModelIndex):
        """
        if index.column() != 0:
            super(NodeNavigationDelagate, self).paint(painter, option, index)
            return

        item = index.model().item(index.row(), index.column())

        margin = 1.0, 1.0
        rect = QtCore.QRectF(
            option.rect.x() + margin[0],
            option.rect.y() + margin[1],
            option.rect.width() - (margin[0] * 2),
            option.rect.height() - (margin[1] * 2),
        )

        painter.save()
        painter.setPen(QtCore.Qt.PenStyle.NoPen)
        painter.setBrush(QtCore.Qt.BrushStyle.NoBrush)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing, True)

        # background.
        bg_color = QtGui.QColor(*ViewerNavEnum.ITEM_COLOR.value)
        itm_color = QtGui.QColor(80, 128, 123)
        if option.state & QtWidgets.QStyle.StateFlag.State_Selected:
            bg_color = bg_color.lighter(120)
            itm_color = QtGui.QColor(*NodeEnum.SELECTED_BORDER_COLOR.value)

        roundness = 2.0
        painter.setBrush(bg_color)
        painter.drawRoundedRect(rect, roundness, roundness)

        if index.row() != 0:
            txt_offset = 8.0
            m = 6.0
            x = rect.left() + 2.0 + m
            y = rect.top() + m + 2
            h = rect.height() - (m * 2) - 2
            painter.setBrush(itm_color)
            for i in range(4):
                itm_rect = QtCore.QRectF(x, y, 1.3, h)
                painter.drawRoundedRect(itm_rect, 1.0, 1.0)
                x += 2.0
                y += 2
                h -= 4
        else:
            txt_offset = 5.0
            x = rect.left() + 4.0
            size = 10.0
            for clr in [QtGui.QColor(0, 0, 0, 80), itm_color]:
                itm_rect = QtCore.QRectF(x, rect.center().y() - (size / 2), size, size)
                painter.setBrush(clr)
                painter.drawRoundedRect(itm_rect, 2.0, 2.0)
                size -= 5.0
                x += 2.5

        # text
        # pen_color = option.palette.text().color()
        pen_color = QtGui.QColor(
            *tuple(map(lambda i, j: i - j, (255, 255, 255), bg_color.getRgb()))
        )
        pen = QtGui.QPen(pen_color, 0.5)
        pen.setCapStyle(QtCore.Qt.PenCapStyle.RoundCap)
        painter.setPen(pen)

        font = painter.font()
        font_metrics = QtGui.QFontMetrics(font)
        item_text = item.text().replace(" ", "_")
        if hasattr(font_metrics, "horizontalAdvance"):
            font_width = font_metrics.horizontalAdvance(item_text)
        else:
            font_width = font_metrics.horizontalAdvance(item_text)
        font_height = font_metrics.height()
        text_rect = QtCore.QRectF(
            rect.center().x() - (font_width / 2) + txt_offset,
            rect.center().y() - (font_height / 2),
            font_width,
            font_height,
        )
        painter.drawText(text_rect, item.text())
        painter.restore()


class NodeNavigationWidget(QtWidgets.QListView):

    navigation_changed = QtCore.Signal(str, list)

    def __init__(self, parent=None):
        super(NodeNavigationWidget, self).__init__(parent)
        self.setSelectionMode(QtWidgets.QListView.SelectionMode.SingleSelection)
        self.setResizeMode(self.ResizeMode.Adjust)
        self.setViewMode(QtWidgets.QListView.ViewMode.ListMode)
        self.setFlow(QtWidgets.QListView.Flow.LeftToRight)
        self.setDragEnabled(False)
        self.setMinimumHeight(20)
        self.setMaximumHeight(36)
        self.setSpacing(0)

        # self.viewport().setAutoFillBackground(False)
        self.setStyleSheet(
            "QListView {{border: 0px;background-color: rgb({0},{1},{2});}}".format(
                *ViewerNavEnum.BACKGROUND_COLOR.value
            )
        )

        self.setItemDelegate(NodeNavigationDelagate(self))
        self.setModel(QtGui.QStandardItemModel())

    def keyPressEvent(self, event):
        event.ignore()

    def mouseReleaseEvent(self, event):
        super(NodeNavigationWidget, self).mouseReleaseEvent(event)
        if not self.selectedIndexes():
            return
        index = self.selectedIndexes()[0]
        rows = reversed(range(1, self.model().rowCount()))
        if index.row() == 0:
            rows = [r for r in rows if r > 0]
        else:
            rows = [r for r in rows if index.row() < r]
        if not rows:
            return
        rm_node_ids = [self.model().item(r, 0).toolTip() for r in rows]
        node_id = self.model().item(index.row(), 0).toolTip()
        [self.model().removeRow(r) for r in rows]
        self.navigation_changed.emit(node_id, rm_node_ids)

    def clear(self):
        self.model().sourceMode().clear()

    def add_label_item(self, label, node_id):
        item = QtGui.QStandardItem(label)
        item.setToolTip(node_id)
        metrics = QtGui.QFontMetrics(item.font())
        if hasattr(metrics, "horizontalAdvance"):
            width = metrics.horizontalAdvance(item.text())
        else:
            width = metrics.horizontalAdvance(item.text())
        width *= 1.5
        item.setSizeHint(QtCore.QSize(width, 20))
        self.model().appendRow(item)
        self.selectionModel().setCurrentIndex(
            self.model().indexFromItem(item),
            QtCore.QItemSelectionModel.SelectionFlag.ClearAndSelect,
        )

    def update_label_item(self, label, node_id):
        rows = reversed(range(self.model().rowCount()))
        for r in rows:
            item = self.model().item(r, 0)
            if item.toolTip() == node_id:
                item.setText(label)

    def remove_label_item(self, node_id):
        rows = reversed(range(1, self.model().rowCount()))
        node_ids = [self.model().item(r, 0).toolTip() for r in rows]
        if node_id not in node_ids:
            return
        index = node_ids.index(node_id)
        if index == 0:
            rows = [r for r in rows if r > 0]
        else:
            rows = [r for r in rows if index < r]
        [self.model().removeRow(r) for r in rows]


if __name__ == "__main__":
    import sys

    def on_nav_changed(selected_id, remove_ids):
        print(selected_id, remove_ids)

    app = QtWidgets.QApplication(sys.argv)

    widget = NodeNavigationWidget()
    widget.navigation_changed.connect(on_nav_changed)

    widget.add_label_item("Close Graph", "root")
    for i in range(1, 5):
        widget.add_label_item("group node {}".format(i), "node_id{}".format(i))
    widget.resize(600, 30)
    widget.show()

    app.exec_()



import cattrs.strategies

from caqtus.types.expression import Expression
from caqtus.utils.serialization import copy_converter
from .transformation import EvaluableOutput, Transformation

converter = copy_converter()


def structure_evaluable_output(data, _) -> EvaluableOutput:
    if isinstance(data, str):
        return Expression(data)
    else:
        return converter.structure(data, Transformation)


def get_converter() -> cattrs.Converter:
    return converter.copy()



#!/usr/bin/python
# -*- coding: utf-8 -*-
from PySide6 import QtWidgets, QtCore, QtGui

from ..constants import MIME_TYPE, URN_SCHEME

TYPE_NODE = QtWidgets.QTreeWidgetItem.UserType + 1
TYPE_CATEGORY = QtWidgets.QTreeWidgetItem.UserType + 2


class _BaseNodeTreeItem(QtWidgets.QTreeWidgetItem):

    def __eq__(self, other):
        """
        Workaround fix for QTreeWidgetItem "operator not implemented error".
        see link: https://bugreports.qt.io/browse/PYSIDE-74
        """
        return id(self) == id(other)


class NodesTreeWidget(QtWidgets.QTreeWidget):
    """
    The :class:`NodeGraphQt.NodesTreeWidget` is a widget for displaying all
    registered nodes from the node graph with this widget a user can create
    nodes by dragging and dropping.

    .. inheritance-diagram:: NodeGraphQt.NodesTreeWidget
        :parts: 1
        :top-classes: PySide2.QtWidgets.QWidget

    .. image:: ../_images/nodes_tree.png
        :width: 300px

    .. code-block:: python
        :linenos:

        from NodeGraphQt import NodeGraph, NodesTreeWidget

        # create node graph.
        graph = NodeGraph()

        # create node tree widget.
        nodes_tree = NodesTreeWidget(parent=None, node_graph=graph)
        nodes_tree.show()

    Args:
        parent (QtWidgets.QWidget): parent of the new widget.
        node_graph (NodeGraphQt.NodeGraph): node graph.
    """

    def __init__(self, parent=None, node_graph=None):
        super(NodesTreeWidget, self).__init__(parent)
        self.setDragDropMode(QtWidgets.QAbstractItemView.DragOnly)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.setHeaderHidden(True)
        self.setWindowTitle("Nodes")

        self._factory = node_graph.node_factory if node_graph else None
        self._custom_labels = {}
        self._category_items = {}

        self._build_tree()

    def __repr__(self):
        return "<{} object at {}>".format(self.__class__.__name__, hex(id(self)))

    def mimeData(self, items):
        node_ids = ["node:{}".format(i.toolTip(0)) for i in items]
        node_urn = URN_SCHEME + ";".join(node_ids)
        mime_data = QtCore.QMimeData()
        mime_data.setData(MIME_TYPE, QtCore.QByteArray(node_urn.encode()))
        return mime_data

    def _build_tree(self):
        """
        Populate the node tree.
        """
        self.clear()
        categories = set()
        node_types = {}
        for name, node_ids in self._factory.names.items():
            for nid in node_ids:
                categories.add(".".join(nid.split(".")[:-1]))
                node_types[nid] = name

        self._category_items = {}
        for category in sorted(categories):
            if category in self._custom_labels.keys():
                label = self._custom_labels[category]
            else:
                label = "{}".format(category)
            cat_item = _BaseNodeTreeItem(self, [label], type=TYPE_CATEGORY)
            cat_item.setFirstColumnSpanned(True)
            cat_item.setFlags(QtCore.Qt.ItemIsEnabled)
            cat_item.setSizeHint(0, QtCore.QSize(100, 26))
            self.addTopLevelItem(cat_item)
            cat_item.setExpanded(True)
            self._category_items[category] = cat_item

        for node_id, node_name in node_types.items():
            category = ".".join(node_id.split(".")[:-1])
            category_item = self._category_items[category]

            item = _BaseNodeTreeItem(category_item, [node_name], type=TYPE_NODE)
            item.setToolTip(0, node_id)
            item.setSizeHint(0, QtCore.QSize(100, 26))

            category_item.addChild(item)

    def _set_node_factory(self, factory):
        """
        Set current node factory.

        Args:
            factory (NodeFactory): node factory.
        """
        self._factory = factory

    def set_category_label(self, category, label):
        """
        Override the label for a node category root item.

        .. image:: ../_images/nodes_tree_category_label.png
            :width: 70%

        Args:
            category (str): node identifier category eg. ``"nodes.widgets"``
            label (str): custom display label. eg. ``"Node Widgets"``
        """
        self._custom_labels[category] = label
        if category in self._category_items:
            item = self._category_items[category]
            item.setText(0, label)

    def update(self):
        """
        Update and refresh the node tree widget.
        """
        self._build_tree()



"""Create tables

Revision ID: 038164d73465
Revises:
Create Date: 2024-08-23 21:21:00.651338

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "038164d73465"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "default_device_configurations",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("device_type", sa.String(length=255), nullable=False),
        sa.Column("content", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("name"),
    )
    op.create_table(
        "parameters",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("content", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("name"),
    )
    op.create_table(
        "path",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("path", sa.String(length=255), nullable=False),
        sa.Column("parent_id", sa.Integer(), nullable=True),
        sa.Column("creation_date", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(["parent_id"], ["path.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_path_id"), "path", ["id"], unique=False)
    op.create_index(op.f("ix_path_parent_id"), "path", ["parent_id"], unique=False)
    op.create_index(op.f("ix_path_path"), "path", ["path"], unique=True)
    op.create_table(
        "sequences",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("path_id", sa.Integer(), nullable=False),
        sa.Column(
            "state",
            sa.Enum(
                "DRAFT",
                "PREPARING",
                "RUNNING",
                "FINISHED",
                "INTERRUPTED",
                "CRASHED",
                name="state",
            ),
            nullable=False,
        ),
        sa.Column("start_time", sa.DateTime(), nullable=True),
        sa.Column("stop_time", sa.DateTime(), nullable=True),
        sa.Column("expected_number_of_shots", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["path_id"], ["path.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_sequences_path_id"), "sequences", ["path_id"], unique=True)
    op.create_table(
        "sequence.device_configurations",
        sa.Column("id_", sa.Integer(), nullable=False),
        sa.Column("sequence_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("device_type", sa.String(length=255), nullable=False),
        sa.Column("content", sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(["sequence_id"], ["sequences.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id_"),
        sa.UniqueConstraint("sequence_id", "name", name="device_configuration"),
    )
    op.create_table(
        "sequence.iteration",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("sequence_id", sa.Integer(), nullable=False),
        sa.Column("content", sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(["sequence_id"], ["sequences.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("sequence_id"),
    )
    op.create_index(
        op.f("ix_sequence.iteration_sequence_id"),
        "sequence.iteration",
        ["sequence_id"],
        unique=False,
    )
    op.create_table(
        "sequence.parameters",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("sequence_id", sa.Integer(), nullable=False),
        sa.Column("content", sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(["sequence_id"], ["sequences.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("sequence_id"),
    )
    op.create_index(
        op.f("ix_sequence.parameters_sequence_id"),
        "sequence.parameters",
        ["sequence_id"],
        unique=False,
    )
    op.create_table(
        "sequence.time_lanes",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("sequence_id", sa.Integer(), nullable=False),
        sa.Column("content", sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(["sequence_id"], ["sequences.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("sequence_id"),
    )
    op.create_index(
        op.f("ix_sequence.time_lanes_sequence_id"),
        "sequence.time_lanes",
        ["sequence_id"],
        unique=False,
    )
    op.create_table(
        "shots",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("sequence_id", sa.Integer(), nullable=False),
        sa.Column("index", sa.Integer(), nullable=False),
        sa.Column("start_time", sa.DateTime(), nullable=False),
        sa.Column("end_time", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(["sequence_id"], ["sequences.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("sequence_id", "index", name="shot_identifier"),
    )
    op.create_index(op.f("ix_shots_index"), "shots", ["index"], unique=False)
    op.create_index(
        op.f("ix_shots_sequence_id"), "shots", ["sequence_id"], unique=False
    )
    op.create_table(
        "shot.data.arrays",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("shot_id", sa.Integer(), nullable=False),
        sa.Column("label", sa.String(length=255), nullable=False),
        sa.Column("dtype", sa.String(length=255), nullable=False),
        sa.Column("shape", sa.JSON(), nullable=True),
        sa.Column("bytes", sa.LargeBinary(), nullable=True),
        sa.ForeignKeyConstraint(["shot_id"], ["shots.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("shot_id", "label", name="array_identifier"),
    )
    op.create_index(
        op.f("ix_shot.data.arrays_shot_id"),
        "shot.data.arrays",
        ["shot_id"],
        unique=False,
    )
    op.create_table(
        "shot.data.structured",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("shot_id", sa.Integer(), nullable=False),
        sa.Column("label", sa.String(length=255), nullable=False),
        sa.Column("content", sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(["shot_id"], ["shots.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("shot_id", "label", name="structured_identifier"),
    )
    op.create_index(
        op.f("ix_shot.data.structured_shot_id"),
        "shot.data.structured",
        ["shot_id"],
        unique=False,
    )
    op.create_table(
        "shot.parameters",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("shot_id", sa.Integer(), nullable=False),
        sa.Column("content", sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(["shot_id"], ["shots.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_shot.parameters_shot_id"), "shot.parameters", ["shot_id"], unique=True
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    raise NotImplementedError
    # ### commands auto generated by Alembic - please adjust! ###
    # op.drop_index(op.f("ix_shot.parameters_shot_id"), table_name="shot.parameters")
    # op.drop_table("shot.parameters")
    # op.drop_index(
    #     op.f("ix_shot.data.structured_shot_id"), table_name="shot.data.structured"
    # )
    # op.drop_table("shot.data.structured")
    # op.drop_index(op.f("ix_shot.data.arrays_shot_id"), table_name="shot.data.arrays")
    # op.drop_table("shot.data.arrays")
    # op.drop_index(op.f("ix_shots_sequence_id"), table_name="shots")
    # op.drop_index(op.f("ix_shots_index"), table_name="shots")
    # op.drop_table("shots")
    # op.drop_index(
    #     op.f("ix_sequence.time_lanes_sequence_id"), table_name="sequence.time_lanes"
    # )
    # op.drop_table("sequence.time_lanes")
    # op.drop_index(
    #     op.f("ix_sequence.parameters_sequence_id"), table_name="sequence.parameters"
    # )
    # op.drop_table("sequence.parameters")
    # op.drop_index(
    #     op.f("ix_sequence.iteration_sequence_id"), table_name="sequence.iteration"
    # )
    # op.drop_table("sequence.iteration")
    # op.drop_table("sequence.device_configurations")
    # op.drop_index(op.f("ix_sequences_path_id"), table_name="sequences")
    # op.drop_table("sequences")
    # op.drop_index(op.f("ix_path_path"), table_name="path")
    # op.drop_index(op.f("ix_path_parent_id"), table_name="path")
    # op.drop_index(op.f("ix_path_id"), table_name="path")
    # op.drop_table("path")
    # op.drop_table("parameters")
    # op.drop_table("default_device_configurations")
    # ### end Alembic commands ###



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'main_window.ui'
##
## Created by: Qt User Interface Compiler version 6.6.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QAction,
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QMainWindow,
    QMenu,
    QMenuBar,
    QSizePolicy,
    QStatusBar,
    QWidget,
)


class Ui_ShotViewerMainWindow(object):
    def setupUi(self, ShotViewerMainWindow):
        if not ShotViewerMainWindow.objectName():
            ShotViewerMainWindow.setObjectName("ShotViewerMainWindow")
        ShotViewerMainWindow.resize(800, 515)
        self.action_save_workspace_as = QAction(ShotViewerMainWindow)
        self.action_save_workspace_as.setObjectName("action_save_workspace_as")
        self.action_load_workspace = QAction(ShotViewerMainWindow)
        self.action_load_workspace.setObjectName("action_load_workspace")
        self.tile_action = QAction(ShotViewerMainWindow)
        self.tile_action.setObjectName("tile_action")
        self.centralwidget = QWidget(ShotViewerMainWindow)
        self.centralwidget.setObjectName("centralwidget")
        ShotViewerMainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(ShotViewerMainWindow)
        self.menubar.setObjectName("menubar")
        self.menubar.setGeometry(QRect(0, 0, 800, 21))
        self.menu_add_view = QMenu(self.menubar)
        self.menu_add_view.setObjectName("menu_add_view")
        self.menuWorkspace = QMenu(self.menubar)
        self.menuWorkspace.setObjectName("menuWorkspace")
        self.menuWindows = QMenu(self.menubar)
        self.menuWindows.setObjectName("menuWindows")
        ShotViewerMainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(ShotViewerMainWindow)
        self.statusbar.setObjectName("statusbar")
        ShotViewerMainWindow.setStatusBar(self.statusbar)

        self.menubar.addAction(self.menu_add_view.menuAction())
        self.menubar.addAction(self.menuWorkspace.menuAction())
        self.menubar.addAction(self.menuWindows.menuAction())
        self.menuWorkspace.addAction(self.action_save_workspace_as)
        self.menuWorkspace.addAction(self.action_load_workspace)
        self.menuWindows.addAction(self.tile_action)

        self.retranslateUi(ShotViewerMainWindow)

        QMetaObject.connectSlotsByName(ShotViewerMainWindow)

    # setupUi

    def retranslateUi(self, ShotViewerMainWindow):
        ShotViewerMainWindow.setWindowTitle(
            QCoreApplication.translate("ShotViewerMainWindow", "MainWindow", None)
        )
        self.action_save_workspace_as.setText(
            QCoreApplication.translate("ShotViewerMainWindow", "Save as...", None)
        )
        self.action_load_workspace.setText(
            QCoreApplication.translate("ShotViewerMainWindow", "Load...", None)
        )
        self.tile_action.setText(
            QCoreApplication.translate("ShotViewerMainWindow", "Tile", None)
        )
        self.menu_add_view.setTitle(
            QCoreApplication.translate("ShotViewerMainWindow", "Add view", None)
        )
        self.menuWorkspace.setTitle(
            QCoreApplication.translate("ShotViewerMainWindow", "Workspace", None)
        )
        self.menuWindows.setTitle(
            QCoreApplication.translate("ShotViewerMainWindow", "Windows", None)
        )

    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'image_view_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.6.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractButton,
    QApplication,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)


class Ui_ImageViewDialog(object):
    def setupUi(self, ImageViewDialog):
        if not ImageViewDialog.objectName():
            ImageViewDialog.setObjectName("ImageViewDialog")
        ImageViewDialog.resize(400, 300)
        ImageViewDialog.setModal(True)
        self.verticalLayout = QVBoxLayout(ImageViewDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.viewNameLabel = QLabel(ImageViewDialog)
        self.viewNameLabel.setObjectName("viewNameLabel")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.viewNameLabel)

        self.view_name_line_edit = QLineEdit(ImageViewDialog)
        self.view_name_line_edit.setObjectName("view_name_line_edit")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.view_name_line_edit)

        self.cameraNameLabel = QLabel(ImageViewDialog)
        self.cameraNameLabel.setObjectName("cameraNameLabel")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.cameraNameLabel)

        self.camera_name_line_edit = QLineEdit(ImageViewDialog)
        self.camera_name_line_edit.setObjectName("camera_name_line_edit")

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.camera_name_line_edit)

        self.imageLabel = QLabel(ImageViewDialog)
        self.imageLabel.setObjectName("imageLabel")

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.imageLabel)

        self.image_line_edit = QLineEdit(ImageViewDialog)
        self.image_line_edit.setObjectName("image_line_edit")

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self.image_line_edit)

        self.backgroundLabel = QLabel(ImageViewDialog)
        self.backgroundLabel.setObjectName("backgroundLabel")

        self.formLayout.setWidget(3, QFormLayout.LabelRole, self.backgroundLabel)

        self.background_line_edit = QLineEdit(ImageViewDialog)
        self.background_line_edit.setObjectName("background_line_edit")

        self.formLayout.setWidget(3, QFormLayout.FieldRole, self.background_line_edit)

        self.verticalLayout.addLayout(self.formLayout)

        self.buttonBox = QDialogButtonBox(ImageViewDialog)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(ImageViewDialog)
        self.buttonBox.accepted.connect(ImageViewDialog.accept)
        self.buttonBox.rejected.connect(ImageViewDialog.reject)

        QMetaObject.connectSlotsByName(ImageViewDialog)

    # setupUi

    def retranslateUi(self, ImageViewDialog):
        ImageViewDialog.setWindowTitle(
            QCoreApplication.translate("ImageViewDialog", "Create image view...", None)
        )
        self.viewNameLabel.setText(
            QCoreApplication.translate("ImageViewDialog", "View name", None)
        )
        self.cameraNameLabel.setText(
            QCoreApplication.translate("ImageViewDialog", "Camera", None)
        )
        self.imageLabel.setText(
            QCoreApplication.translate("ImageViewDialog", "Image", None)
        )
        self.backgroundLabel.setText(
            QCoreApplication.translate("ImageViewDialog", "Background", None)
        )
        self.background_line_edit.setPlaceholderText(
            QCoreApplication.translate("ImageViewDialog", "None", None)
        )

    # retranslateUi



"""Provides an implementation of experiment sessions using SQL databases."""

from ._session_maker import (
    PostgreSQLExperimentSessionMaker,
    PostgreSQLConfig,
)

__all__ = [
    "PostgreSQLExperimentSessionMaker",
    "PostgreSQLConfig",
]



from typing import NewType

DeviceParameter = NewType("DeviceParameter", str)
"""A string that represents the name of a device parameter.

The string should have python variable naming conventions.
"""



from collections.abc import Callable
from typing import Protocol

from caqtus.device import DeviceConfiguration, Device, DeviceController
from caqtus.device.configuration import DeviceServerName
from caqtus.device.remote import DeviceProxy, RPCConfiguration
from caqtus.shot_compilation import DeviceCompiler


class DeviceManagerExtensionProtocol(Protocol):
    def get_device_compiler_type(
        self, device_configuration: DeviceConfiguration
    ) -> type[DeviceCompiler]: ...

    def get_device_type(
        self, device_configuration: DeviceConfiguration
    ) -> Callable[..., Device]: ...

    def get_device_controller_type(
        self, device_configuration: DeviceConfiguration
    ) -> type[DeviceController]: ...

    def get_proxy_type(
        self, device_configuration: DeviceConfiguration
    ) -> type[DeviceProxy]:
        """Returns the type used to create proxies for the given device configuration."""

        ...

    def get_device_server_config(self, server: DeviceServerName) -> RPCConfiguration:
        """Returns the configuration for the given device server.

        raises:
            KeyError: If no configuration is found for the given device server.
        """

        ...



import functools

import numpy as np

from ._instructions import SequencerInstruction, Pattern, Concatenated, Repeated
from ._ramp import Ramp


def convert_to_change_arrays(
    sequence: SequencerInstruction,
) -> tuple[np.ndarray, np.ndarray]:
    """Convert a sequence to arrays of change times and values.

    Args:
        sequence: The sequence to convert.

    Returns:
        Two arrays, the first containing times indexes at which the values of the
        sequence change, and the second containing the corresponding values for each
        change.
        The time array will have dtype np.int64 and the value array will have
        the same dtype as the sequence passed in argument.
    """

    times, values = _convert_to_change_arrays(sequence)
    return np.concatenate([times, [len(sequence)]], axis=0), np.concatenate(
        [values, [values[-1]]], axis=0
    )


@functools.singledispatch
def _convert_to_change_arrays(
    sequence: SequencerInstruction,
) -> tuple[np.ndarray, np.ndarray]:
    raise NotImplementedError(f"Cannot convert {type(sequence)} to time arrays")


@_convert_to_change_arrays.register
def _convert_pattern(sequence: Pattern) -> tuple[np.ndarray, np.ndarray]:
    times = np.arange(len(sequence), dtype=np.int64)
    values = sequence.array

    return times, values


@_convert_to_change_arrays.register
def _convert_ramp(sequence: Ramp) -> tuple[np.ndarray, np.ndarray]:
    return _convert_pattern(sequence.to_pattern())


@_convert_to_change_arrays.register
def _(sequence: Concatenated) -> tuple[np.ndarray, np.ndarray]:
    time_arrays = []
    value_arrays = []

    cumulated_start = 0

    for instruction in sequence.instructions:
        times, values = _convert_to_change_arrays(instruction)
        time_arrays.append(times + cumulated_start)
        value_arrays.append(values)
        cumulated_start += len(instruction)

    return np.concatenate(time_arrays), np.concatenate(value_arrays)


@_convert_to_change_arrays.register
def _(sequence: Repeated) -> tuple[np.ndarray, np.ndarray]:
    time_array, value_array = _convert_to_change_arrays(sequence.instruction)
    if len(time_array) == 1:
        return np.array([0], dtype=np.int64), value_array
    start_times = np.arange(sequence.repetitions, dtype=np.int64) * len(
        sequence.instruction
    )
    time_arrays = time_array + start_times[:, np.newaxis]
    return np.concatenate(time_arrays), np.tile(value_array, sequence.repetitions)



from __future__ import annotations

from typing import TypeAlias

import numpy as np

#: A type alias for numpy arrays.
type Array = np.ndarray


StructuredData: TypeAlias = (
    dict[str, "StructuredData"]
    | list["StructuredData"]
    | float
    | int
    | str
    | bool
    | None
)
"""A recursive union of basic types.

It can be used to represent complex data, possibly nested.
"""

type Data = Array | StructuredData
"""A sum type of structured data and arrays.

Only objects that are instances of this type can be saved for a shot.

Note that it is not possible to have structured data containing arrays.
"""



from PySide6 import QtWidgets, QtGui

from ..constants import NodeEnum, ViewerEnum, ViewerNavEnum

from .viewer_nav import NodeNavigationWidget


class NodeGraphWidget(QtWidgets.QTabWidget):

    def __init__(self, parent=None):
        super(NodeGraphWidget, self).__init__(parent)
        self.setTabsClosable(True)
        self.setTabBarAutoHide(True)
        bg_color = QtGui.QColor(*ViewerEnum.BACKGROUND_COLOR.value).darker(120).getRgb()
        text_color = tuple(map(lambda i, j: i - j, (255, 255, 255), bg_color))
        style_dict = {
            "QWidget": {
                "background-color": "rgb({0},{1},{2})".format(
                    *ViewerEnum.BACKGROUND_COLOR.value
                ),
            },
            "QTabWidget::pane": {
                "background": "rgb({0},{1},{2})".format(
                    *ViewerEnum.BACKGROUND_COLOR.value
                ),
                "border": "0px",
                "border-top": "0px solid rgb({0},{1},{2})".format(*bg_color),
            },
            "QTabBar::tab": {
                "background": "rgb({0},{1},{2})".format(*bg_color),
                "border": "0px solid black",
                "color": "rgba({0},{1},{2},30)".format(*text_color),
                "min-width": "10px",
                "padding": "10px 20px",
            },
            "QTabBar::tab:selected": {
                "color": "rgb({0},{1},{2})".format(*text_color),
                "background": "rgb({0},{1},{2})".format(
                    *ViewerNavEnum.BACKGROUND_COLOR.value
                ),
                "border-top": "1px solid rgb({0},{1},{2})".format(
                    *NodeEnum.SELECTED_BORDER_COLOR.value
                ),
            },
            "QTabBar::tab:hover": {
                "color": "rgb({0},{1},{2})".format(*text_color),
                "border-top": "1px solid rgb({0},{1},{2})".format(
                    *NodeEnum.SELECTED_BORDER_COLOR.value
                ),
            },
        }
        stylesheet = ""
        for css_class, css in style_dict.items():
            style = "{} {{\n".format(css_class)
            for elm_name, elm_val in css.items():
                style += "  {}:{};\n".format(elm_name, elm_val)
            style += "}\n"
            stylesheet += style
        self.setStyleSheet(stylesheet)

    def add_viewer(self, viewer, name, node_id):
        self.addTab(viewer, name)
        index = self.indexOf(viewer)
        self.setTabToolTip(index, node_id)
        self.setCurrentIndex(index)

    def remove_viewer(self, viewer):
        index = self.indexOf(viewer)
        self.removeTab(index)


class SubGraphWidget(QtWidgets.QWidget):

    def __init__(self, parent=None, graph=None):
        super(SubGraphWidget, self).__init__(parent)
        self._graph = graph
        self._navigator = NodeNavigationWidget()
        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(1)
        self._layout.addWidget(self._navigator)

        self._viewer_widgets = {}
        self._viewer_current = None

    @property
    def navigator(self):
        return self._navigator

    def add_viewer(self, viewer, name, node_id):
        if viewer in self._viewer_widgets:
            return

        if self._viewer_current:
            self.hide_viewer(self._viewer_current)

        self._navigator.add_label_item(name, node_id)
        self._layout.addWidget(viewer)
        self._viewer_widgets[viewer] = node_id
        self._viewer_current = viewer
        self._viewer_current.show()

    def remove_viewer(self, viewer=None):
        if viewer is None and self._viewer_current:
            viewer = self._viewer_current
        node_id = self._viewer_widgets.pop(viewer)
        self._navigator.remove_label_item(node_id)
        self._layout.removeWidget(viewer)
        viewer.deleteLater()

    def hide_viewer(self, viewer):
        self._layout.removeWidget(viewer)
        viewer.hide()

    def show_viewer(self, viewer):
        if viewer == self._viewer_current:
            self._viewer_current.show()
            return
        if viewer in self._viewer_widgets:
            if self._viewer_current:
                self.hide_viewer(self._viewer_current)
            self._layout.addWidget(viewer)
            self._viewer_current = viewer
            self._viewer_current.show()



import functools
from collections.abc import Iterable, Callable, Generator
from typing import assert_never, TypeVar

import caqtus.formatter as fmt
from caqtus.types.iteration import (
    Step,
    ArangeLoop,
    ExecuteShot,
    VariableDeclaration,
    LinspaceLoop,
)
from caqtus.types.parameter import is_parameter, Parameter, ParameterNamespace
from caqtus.types.recoverable_exceptions import InvalidTypeError
from .shots_manager import ShotScheduler
from .step_context import StepContext

S = TypeVar("S", bound=Step)


def wrap_error(
    function: Callable[[S, StepContext], Generator[StepContext, None, StepContext]]
) -> Callable[[S, StepContext], Generator[StepContext, None, StepContext]]:
    """Wrap a function that evaluates a step to raise nicer errors for the user."""

    @functools.wraps(function)
    def wrapper(step: S, context: StepContext):
        try:
            return function(step, context)
        except Exception as e:
            raise StepEvaluationError(f"Error while evaluating step <{step}>") from e

    return wrapper


async def execute_steps(
    steps: Iterable[Step],
    initial_context: StepContext[Parameter],
    shot_scheduler: ShotScheduler,
):
    """Execute a sequence of steps on the experiment.

    This method will recursively execute each step in the sequence passed as
    argument and scheduling the shots when encountering a shot step.
    """

    for context in walk_steps(steps, initial_context):
        await shot_scheduler.schedule_shot(context.variables)


def walk_steps(
    steps: Iterable[Step], initial_context: StepContext
) -> Iterable[StepContext]:
    """Yields the context for each shot defined by the steps.

    This function will recursively evaluate each step in the sequence passed as
    argument.
    Before executing the sequence, an empty context is initialized.
    The context holds the value of the parameters at a given point in the sequence.
    Each step has the possibility to update the context with new values.
    """

    context = initial_context

    for step in steps:
        context = yield from walk_step(step, context)


@functools.singledispatch
@wrap_error
def walk_step(
    step: Step, context: StepContext
) -> Generator[StepContext, None, StepContext]:
    """Iterates over the steps of a sequence.

    Args:
        step: the step of the sequence currently executed
        context: Contains the values of the variables before this step.

    Yields:
        The context for every shot encountered while walking the steps.

    Returns:
        The context after the step passed in argument has been executed.
    """

    return assert_never(step)  # type: ignore


# noinspection PyUnreachableCode
@walk_step.register
@wrap_error
def _(
    declaration: VariableDeclaration,
    context: StepContext,
) -> Generator[StepContext, None, StepContext]:
    """Execute a VariableDeclaration step.

    This step updates the context passed with the value of the variable declared.
    """

    value = declaration.value.evaluate(context.variables.dict())
    if not is_parameter(value):
        raise InvalidTypeError(
            f"{fmt.expression(declaration.value)}> does not evaluate to a parameter, "
            f"but to {fmt.type_(type(value))}.",
        )
    return context.update_variable(declaration.variable, value)

    # This code is unreachable, but it is kept here to make the function a generator.
    if False:
        yield context


@walk_step.register
@wrap_error
def _(
    arange_loop: ArangeLoop,
    context: StepContext,
) -> Generator[StepContext, None, StepContext]:
    """Loop over a variable in a numpy arange like loop.

    This function will loop over the variable defined in the arange loop and execute
    the sub steps for each value of the variable.

    Returns:
        A new context object containing the value of the arange loop variable after
        the last iteration, plus the values of the variables defined in the sub
        steps.
    """

    for value in arange_loop.loop_values(context.variables.dict()):
        context = context.update_variable(arange_loop.variable, value)
        for step in arange_loop.sub_steps:
            context = yield from walk_step(step, context)
    return context


@walk_step.register
@wrap_error
def _(
    linspace_loop: LinspaceLoop,
    context: StepContext,
) -> Generator[StepContext, None, StepContext]:
    """Loop over a variable in a numpy linspace like loop.

    This function will loop over the variable defined in the linspace loop and
    execute the sub steps for each value of the variable.

    Returns:
        A new context object containing the value of the linspace loop variable
        after the last iteration, plus the values of the variables defined in the
        sub steps.
    """

    for value in linspace_loop.loop_values(context.variables.dict()):
        context = context.update_variable(linspace_loop.variable, value)
        for step in linspace_loop.sub_steps:
            context = yield from walk_step(step, context)
    return context


@walk_step.register
@wrap_error
def _(
    shot: ExecuteShot, context: StepContext
) -> Generator[StepContext, None, StepContext]:
    """Schedule a shot to be run.

    This function schedule to run a shot on the experiment with the parameters
    defined in the context at this point.

    Returns:
        The context passed as argument unchanged.
    """

    yield context
    return context


class StepEvaluationError(Exception):
    pass


def evaluate_initial_context(parameters: ParameterNamespace) -> StepContext:
    """Evaluate the initial context of the sequence from the parameters."""

    flat_parameters = parameters.flatten()

    context = StepContext[Parameter]()

    for name, expression in flat_parameters:
        value = expression.evaluate({})
        if not is_parameter(value):
            raise InvalidTypeError(
                f"{fmt.expression(value)} for {fmt.shot_param(name)} does not evaluate "
                f"to a parameter, but to {fmt.type_(type(value))}",
            )
        context = context.update_variable(name, value)

    return context



import contextlib
import functools
from datetime import datetime
from typing import Callable, Concatenate, TypeVar, ParamSpec, Mapping, Optional, Self

import anyio.to_thread
import attrs
from returns.result import Result
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session

from caqtus.types.data import DataLabel, Data
from caqtus.types.iteration import IterationConfiguration
from caqtus.types.parameter import Parameter, ParameterNamespace
from caqtus.types.timelane import TimeLanes
from caqtus.types.variable_name import DottedVariableName
from ._experiment_session import _get_global_parameters, _set_global_parameters
from ._path_hierarchy import _does_path_exists, _get_children, _get_path_creation_date
from ._sequence_collection import (
    _get_stats,
    _is_sequence,
    _get_sequence_global_parameters,
    _get_time_lanes,
    _get_iteration_configuration,
    _get_shots,
    _get_shot_parameters,
    _get_shot_end_time,
    _get_shot_start_time,
    _get_shot_data_by_label,
    _get_all_shot_data,
    _get_exceptions,
)
from ._serializer import SerializerProtocol
from .._exception_summary import TracebackSummary
from .._experiment_session import ExperimentSessionNotActiveError
from .._path import PureSequencePath
from .._path_hierarchy import PathNotFoundError, PathIsRootError
from .._sequence_collection import (
    PathIsSequenceError,
    SequenceStats,
    PathIsNotSequenceError,
    PureShot,
    SequenceNotCrashedError,
)
from ..async_session import (
    AsyncExperimentSession,
    AsyncPathHierarchy,
    AsyncSequenceCollection,
)

_T = TypeVar("_T")
_P = ParamSpec("_P")


class AsyncSQLExperimentSession(AsyncExperimentSession):
    def __init__(
        self,
        async_session_context: contextlib.AbstractAsyncContextManager[AsyncSession],
        serializer: SerializerProtocol,
    ):
        self._async_session_context = async_session_context
        self._async_session: Optional[AsyncSession] = None
        self.paths = AsyncSQLPathHierarchy(parent_session=self)
        self.sequences = AsyncSQLSequenceCollection(
            parent_session=self, serializer=serializer
        )

    async def __aenter__(self) -> Self:
        if self._async_session is not None:
            error = RuntimeError("Session has already been activated")
            error.add_note(
                "You cannot reactivate a session, you must create a new one."
            )
        self._async_session = await self._async_session_context.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self._async_session_context.__aexit__(exc_type, exc_val, exc_tb)

    async def get_global_parameters(self) -> ParameterNamespace:
        return await self._run_sync(_get_global_parameters)

    async def set_global_parameters(self, parameters: ParameterNamespace) -> None:
        return await self._run_sync(_set_global_parameters, parameters)

    async def _run_sync(
        self,
        fun: Callable[Concatenate[Session, _P], _T],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _T:
        return await self._session().run_sync(fun, *args, **kwargs)

    def _session(self) -> AsyncSession:
        if self._async_session is None:
            raise ExperimentSessionNotActiveError(
                "Experiment session was not activated"
            )
        return self._async_session


class ThreadedAsyncSQLExperimentSession(AsyncSQLExperimentSession):
    def __init__(
        self,
        session_context: contextlib.AbstractContextManager[Session],
        serializer: SerializerProtocol,
    ):
        self._session_context = session_context
        self._session: Optional[Session] = None

        self.paths = AsyncSQLPathHierarchy(parent_session=self)
        self.sequences = AsyncSQLSequenceCollection(
            parent_session=self, serializer=serializer
        )

    async def __aenter__(self) -> Self:
        if self._session is not None:
            raise RuntimeError("Session is already active")
        self._session = await anyio.to_thread.run_sync(self._session_context.__enter__)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await anyio.to_thread.run_sync(
            self._session_context.__exit__, exc_type, exc_val, exc_tb
        )

    async def _run_sync(
        self,
        fun: Callable[Concatenate[Session, _P], _T],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _T:
        if self._session is None:
            raise ExperimentSessionNotActiveError(
                "Experiment session was not activated"
            )
        wrapped = functools.partial(fun, self._session, *args, **kwargs)
        return await anyio.to_thread.run_sync(wrapped)


@attrs.frozen
class AsyncSQLPathHierarchy(AsyncPathHierarchy):
    parent_session: AsyncSQLExperimentSession

    async def does_path_exists(self, path: PureSequencePath) -> bool:
        return await self._run_sync(_does_path_exists, path)

    async def get_children(
        self, path: PureSequencePath
    ) -> Result[set[PureSequencePath], PathNotFoundError | PathIsSequenceError]:
        return await self._run_sync(_get_children, path)

    async def get_path_creation_date(
        self, path: PureSequencePath
    ) -> Result[datetime, PathNotFoundError | PathIsRootError]:
        return await self._run_sync(_get_path_creation_date, path)

    async def _run_sync(
        self,
        fun: Callable[Concatenate[Session, _P], _T],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _T:
        return await self.parent_session._run_sync(fun, *args, **kwargs)


@attrs.frozen
class AsyncSQLSequenceCollection(AsyncSequenceCollection):
    parent_session: AsyncSQLExperimentSession
    serializer: SerializerProtocol

    async def is_sequence(
        self, path: PureSequencePath
    ) -> Result[bool, PathNotFoundError]:
        return await self._run_sync(_is_sequence, path)

    async def get_stats(
        self, path: PureSequencePath
    ) -> Result[SequenceStats, PathNotFoundError | PathIsNotSequenceError]:
        return await self._run_sync(_get_stats, path)

    async def get_traceback_summary(self, path: PureSequencePath) -> Result[
        Optional[TracebackSummary],
        PathNotFoundError | PathIsNotSequenceError | SequenceNotCrashedError,
    ]:
        return await self._run_sync(_get_exceptions, path)

    async def get_time_lanes(self, path: PureSequencePath) -> TimeLanes:
        return await self._run_sync(_get_time_lanes, path, self.serializer)

    async def get_global_parameters(self, path: PureSequencePath) -> ParameterNamespace:
        return await self._run_sync(_get_sequence_global_parameters, path)

    async def get_iteration_configuration(
        self, path: PureSequencePath
    ) -> IterationConfiguration:
        return await self._run_sync(_get_iteration_configuration, path, self.serializer)

    async def get_shots(
        self, path: PureSequencePath
    ) -> Result[list[PureShot], PathNotFoundError | PathIsNotSequenceError]:
        return await self._run_sync(_get_shots, path)

    async def get_shot_parameters(
        self, path: PureSequencePath, shot_index: int
    ) -> Mapping[DottedVariableName, Parameter]:
        return await self._run_sync(_get_shot_parameters, path, shot_index)

    async def get_all_shot_data(
        self, path: PureSequencePath, shot_index: int
    ) -> Mapping[DataLabel, Data]:
        return await self._run_sync(_get_all_shot_data, path, shot_index)

    async def get_shot_data_by_label(
        self, path: PureSequencePath, shot_index: int, data_label: DataLabel
    ) -> Data:
        return await self._run_sync(
            _get_shot_data_by_label, path, shot_index, data_label
        )

    async def get_shot_start_time(
        self, path: PureSequencePath, shot_index: int
    ) -> datetime:
        return await self._run_sync(_get_shot_start_time, path, shot_index)

    async def get_shot_end_time(
        self, path: PureSequencePath, shot_index: int
    ) -> datetime:
        return await self._run_sync(_get_shot_end_time, path, shot_index)

    async def _run_sync(
        self,
        fun: Callable[Concatenate[Session, _P], _T],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _T:
        return await self.parent_session._run_sync(fun, *args, **kwargs)



from __future__ import annotations

import sqlalchemy
from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column

from ._table_base import Base


# This table stores parameters that can be shared across sequences.
# Each set of parameters is identified by a unique name and stores a JSON object
# representing a ParameterNamespace.
# At the moment, there is only one row in this table, which stores the so-called
# "global" parameters.
class SQLParameters(Base):
    __tablename__ = "parameters"

    name: Mapped[str] = mapped_column(String(255), primary_key=True)
    content = mapped_column(sqlalchemy.types.JSON)



"""Contains utilities to generate complex values for the output of a device.

It is convenient to transform input values given by the user into output values that
a device should generate.
This module contains classes that can be used to construct complex tree structures that
represent user defined transformations.
"""

import functools

import cattrs.strategies

from ._converter import converter, structure_evaluable_output
from ._output_mapping import LinearInterpolation
from .transformation import Transformation, evaluate, EvaluableOutput

cattrs.strategies.include_subclasses(
    Transformation,
    converter=converter,
    union_strategy=functools.partial(
        cattrs.strategies.configure_tagged_union, tag_name="type"
    ),
)
converter.register_structure_hook(EvaluableOutput, structure_evaluable_output)

__all__ = [
    "Transformation",
    "LinearInterpolation",
    "evaluate",
    "EvaluableOutput",
    "converter",
]



import functools
from collections.abc import Iterable
from typing import Mapping, Any, Optional

import attrs
import numpy as np

# TODO: Can remove tblib support once the experiment manager runs in a single process
import tblib.pickling_support

from caqtus.device import DeviceName, DeviceParameter
from caqtus.shot_compilation import SequenceContext, ShotContext
from caqtus.shot_compilation.lane_compilers.timing import number_ticks, ns
from caqtus.types.recoverable_exceptions import InvalidValueError
from caqtus.types.units import Unit, InvalidDimensionalityError, dimensionless
from caqtus.types.units.base import is_in_base_units, base_units
from caqtus.types.variable_name import DottedVariableName
from .._time_step import TimeStep
from ..channel_commands import ChannelOutput
from ..channel_commands import DimensionedSeries
from ..channel_commands._channel_sources._trigger_compiler import (
    TriggerableDeviceCompiler,
)
from ..configuration import DigitalChannelConfiguration, AnalogChannelConfiguration
from ..configuration import (
    SequencerConfiguration,
)
from ..instructions import (
    with_name,
    stack_instructions,
    SequencerInstruction,
    Pattern,
    Ramp,
    Concatenated,
    concatenate,
    Repeated,
)
from ..trigger import ExternalClockOnChange, ExternalTriggerStart, SoftwareTrigger


class SequencerCompiler(TriggerableDeviceCompiler):
    """Compile parameters for a sequencer device."""

    def __init__(self, device_name: DeviceName, sequence_context: SequenceContext):
        super().__init__(device_name, sequence_context)
        configuration = sequence_context.get_device_configuration(device_name)
        if not isinstance(configuration, SequencerConfiguration):
            raise TypeError(
                f"Expected a sequencer configuration for device {device_name}, got "
                f"{type(configuration)}"
            )
        self.__configuration = configuration
        self.__device_name = device_name

    def compile_initialization_parameters(self) -> Mapping[DeviceParameter, Any]:
        """Compile the parameters needed to initialize the sequencer.

        Returns:
            A dictionary with the following keys:

            * 'time_step': The time step of the sequencer, in ns.
            * 'trigger': The trigger configuration of the sequencer.
        """

        # TODO: raise DeviceNotUsedException if the sequencer is not used for the
        #  current sequence
        return {
            DeviceParameter("time_step"): self.__configuration.time_step,
            DeviceParameter("trigger"): self.__configuration.trigger,
        }

    def compile_shot_parameters(
        self,
        shot_context: ShotContext,
    ) -> Mapping[str, Any]:
        """Evaluates the output for each channel of the sequencer.

        Returns:
            A dictionary with the following key:

            * 'sequence': A :class:`SequencerInstruction` that contains the
              instructions for all the channels of the sequencer.
        """

        instructions = {}

        for channel_index, channel in enumerate(self.__configuration.channels):
            output = channel.output
            if isinstance(channel, DigitalChannelConfiguration):
                dtype = np.dtype(np.bool_)
                units = None
            elif isinstance(channel, AnalogChannelConfiguration):
                dtype = np.dtype(np.float64)
                units = base_units(Unit(channel.output_unit))
            else:
                raise TypeError(
                    f"Expected a digital or analog channel configuration, got "
                    f"{type(channel)}"
                )
            instructions[f"ch {channel_index}"] = InstructionCompilationParameters(
                description=channel.description,
                output=output,
                dtype=dtype,
                units=units,
            )

        stacked = compile_parallel_instructions(
            instructions,
            self.__configuration.time_step,
            shot_context,
        )

        return {"sequence": stacked}

    def compute_trigger(
        self, sequencer_time_step: TimeStep, shot_context: ShotContext
    ) -> SequencerInstruction[np.bool_]:
        length = number_ticks(
            0, shot_context.get_shot_duration(), sequencer_time_step * ns
        )

        if isinstance(self.__configuration.trigger, ExternalClockOnChange):
            single_clock_pulse = get_master_clock_pulse(
                self.__configuration.time_step, sequencer_time_step
            )
            slave_parameters = shot_context.get_shot_parameters(self.__device_name)
            slave_instruction = slave_parameters["sequence"]
            instruction = get_adaptive_clock(slave_instruction, single_clock_pulse)[
                :length
            ]
            return instruction
        elif isinstance(self.__configuration.trigger, ExternalTriggerStart):
            return super().compute_trigger(sequencer_time_step, shot_context)
        elif isinstance(self.__configuration.trigger, SoftwareTrigger):
            raise InvalidValueError(
                "Can't generate a trigger for a sequencer that is software triggered"
            )
        else:
            raise NotImplementedError(
                f"Can't generate trigger for {self.__configuration.trigger}"
            )


@attrs.frozen
class InstructionCompilationParameters:
    """Specify how to evaluate an instruction for a channel.

    Attributes:
        description: A human-readable description of the channel.

            This is used to identify the channel in error messages.

        output: The output of the channel.

        dtype: The dtype in which the instruction will be converted once the output has
            been evaluated.

        units: The units in which the output of the channel is expressed.

            The units must be expressed in the base units.
            If the values are dimensionless, the units must be None.
    """

    description: str = attrs.field(converter=str)
    output: ChannelOutput = attrs.field(
        validator=attrs.validators.instance_of(ChannelOutput),
    )
    dtype: np.dtype = attrs.field(validator=attrs.validators.instance_of(np.dtype))
    units: Optional[Unit] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(Unit))
    )

    @units.validator  # type: ignore
    def _validate_units(self, _, units: Optional[Unit]):
        if units is not None:
            if not is_in_base_units(units):
                raise ValueError(
                    f"Unit {units} is not expressed in the base units of the registry."
                )
            if units.is_compatible_with(dimensionless):
                raise ValueError(f"Unit {units} is dimensionless and must be None.")


def compile_parallel_instructions(
    instructions: Mapping[str, InstructionCompilationParameters],
    time_step: TimeStep,
    shot_context: ShotContext,
) -> SequencerInstruction:
    """Evaluates and merges the output for different channels.

    Args:
        instructions: A mapping that indicates how to evaluate individual instructions.
        time_step: The time step used to evaluate the instructions.
        shot_context: The context of the shot.

    Returns:
        A :class:`SequencerInstruction` with multiple fields, each corresponding to the
        instruction passed in the `instructions` argument.
    """

    max_advance, max_delay = _find_max_advance_and_delays(
        [instruction.output for instruction in instructions.values()],
        time_step,
        shot_context.get_variables(),
    )

    channel_instructions = []
    exceptions = []
    for instruction_label, instruction in instructions.items():
        try:
            output_series = instruction.output.evaluate(
                time_step,
                max_advance,
                max_delay,
                shot_context,
            )
            instruction = _convert_series_to_instruction(output_series, instruction)
            channel_instructions.append(with_name(instruction, instruction_label))
        except Exception as e:
            try:
                raise ChannelCompilationError(
                    f"Error occurred when evaluating output for "
                    f"'{instruction.description}'"
                ) from e
            except ChannelCompilationError as channel_error:
                exceptions.append(channel_error)
    if exceptions:
        raise SequencerCompilationError(
            f"Errors occurred when evaluating outputs",
            exceptions,
        )
    stacked = stack_instructions(*channel_instructions)
    return stacked


def _find_max_advance_and_delays(
    outputs: Iterable[ChannelOutput],
    time_step: TimeStep,
    variables: Mapping[DottedVariableName, Any],
) -> tuple[int, int]:
    advances_and_delays = [
        output.evaluate_max_advance_and_delay(time_step, variables)
        for output in outputs
    ]
    advances, delays = zip(*advances_and_delays)
    return max(advances), max(delays)


def get_master_clock_pulse(
    slave_time_step: TimeStep, master_time_step: TimeStep
) -> SequencerInstruction[np.bool_]:
    _, high, low = high_low_clicks(slave_time_step, master_time_step)
    single_clock_pulse = Pattern([True]) * high + Pattern([False]) * low
    assert len(single_clock_pulse) * master_time_step == slave_time_step
    return single_clock_pulse


def high_low_clicks(
    slave_time_step: TimeStep, master_timestep: TimeStep
) -> tuple[int, int, int]:
    """Return the number of steps the master sequencer must be high then low to
    produce a clock pulse for the slave sequencer.

    Returns:
        A tuple with its first element being the number of master steps that constitute
        a full slave clock cycle, the second element being the number of master steps
        for which the master must be high and the third element being the number of
        master steps for which the master must be low.
        The first element is the sum of the second and third elements.
    """

    if not slave_time_step >= 2 * master_timestep:
        raise InvalidValueError(
            "Slave time step must be at least twice the master sequencer time step"
        )
    div_decimal, mod = divmod(slave_time_step, master_timestep)
    if not mod == 0:
        raise InvalidValueError(
            "Slave time step must be an integer multiple of the master sequencer time "
            "step"
        )
    div, denominator = div_decimal.as_integer_ratio()
    assert denominator == 1
    if div % 2 == 0:
        return div, div // 2, div // 2
    else:
        return div, div // 2 + 1, div // 2


@functools.singledispatch
def get_adaptive_clock(
    slave_instruction: SequencerInstruction, clock_pulse: SequencerInstruction
) -> SequencerInstruction:
    """Generates a clock signal for a slave instruction."""

    raise NotImplementedError(
        f"Don't know how to generate a clock for an instruction of type "
        f"{type(slave_instruction)}"
    )


@get_adaptive_clock.register
def _(
    target_sequence: Pattern | Ramp, clock_pulse: SequencerInstruction
) -> SequencerInstruction:
    return clock_pulse * len(target_sequence)


@get_adaptive_clock.register
def _(
    target_sequence: Concatenated, clock_pulse: SequencerInstruction
) -> SequencerInstruction:
    return concatenate(
        *(
            get_adaptive_clock(sequence, clock_pulse)
            for sequence in target_sequence.instructions
        )
    )


@get_adaptive_clock.register
def _(
    target_sequence: Repeated, clock_pulse: SequencerInstruction
) -> SequencerInstruction:
    if len(target_sequence.instruction) == 1:
        return clock_pulse + Pattern([False]) * (
            (len(target_sequence) - 1) * len(clock_pulse)
        )
    else:
        raise NotImplementedError(
            "Only one instruction is supported in a repeat block at the moment"
        )


@tblib.pickling_support.install
class SequencerCompilationError(ExceptionGroup):
    pass


@tblib.pickling_support.install
class ChannelCompilationError(Exception):
    pass


def _convert_series_to_instruction(
    series: DimensionedSeries, instruction: InstructionCompilationParameters
) -> SequencerInstruction:
    if instruction.units != series.units:
        raise InvalidDimensionalityError(
            f"Instruction {instruction.description} output has units {series.units}, "
            f"expected {instruction.units}"
        )
    return series.values.as_type(instruction.dtype)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
**NodeGraphQt** is a node graph framework that can be implemented and re purposed
into applications that supports **PySide2**.

project: https://github.com/jchanvfx/NodeGraphQt
documentation: https://jchanvfx.github.io/NodeGraphQt/api/html/index.html

example code:

.. code-block:: python
    :linenos:

    from NodeGraphQt import QtWidgets, NodeGraph, BaseNode


    class MyNode(BaseNode):

        __identifier__ = 'io.github.jchanvfx'
        NODE_NAME = 'My Node'

        def __init__(self):
            super(MyNode, self).__init__()
            self.add_input('foo', color=(180, 80, 0))
            self.add_output('bar')

    if __name__ == '__main__':
        app = QtWidgets.QApplication([])
        graph = NodeGraph()

        graph.register_node(BaseNode)
        graph.register_node(BackdropNode)

        backdrop = graph.create_node('nodeGraphQt.nodes.Backdrop', name='Backdrop')
        node_a = graph.create_node('io.github.jchanvfx.MyNode', name='Node A')
        node_b = graph.create_node('io.github.jchanvfx.MyNode', name='Node B', color='#5b162f')

        node_a.set_input(0, node_b.output(0))

        viewer = graph.viewer()
        viewer.show()

        app.exec_()
"""
from .pkg_info import __version__ as VERSION
from .pkg_info import __license__ as LICENSE

# node graph
from .base.graph import NodeGraph, SubGraph
from .base.menu import NodesMenu, NodeGraphMenu, NodeGraphCommand

# nodes & ports
from .base.port import Port
from .base.node import NodeObject
from .nodes.base_node import BaseNode
from .nodes.base_node_circle import BaseNodeCircle
from .nodes.backdrop_node import BackdropNode
from .nodes.group_node import GroupNode

# widgets
from .widgets.node_widgets import NodeBaseWidget
from .custom_widgets.nodes_tree import NodesTreeWidget
from .custom_widgets.nodes_palette import NodesPaletteWidget
from .custom_widgets.properties_bin.node_property_widgets import (
    NodePropEditorWidget,
    PropertiesBinWidget,
)


__version__ = VERSION
__all__ = [
    "BackdropNode",
    "BaseNode",
    "BaseNodeCircle",
    "GroupNode",
    "LICENSE",
    "NodeBaseWidget",
    "NodeGraph",
    "NodeGraphCommand",
    "NodeGraphMenu",
    "NodeObject",
    "NodesPaletteWidget",
    "NodePropEditorWidget",
    "NodesTreeWidget",
    "NodesMenu",
    "Port",
    "PropertiesBinWidget",
    "SubGraph",
    "VERSION",
    "constants",
    "custom_widgets",
]



from collections.abc import Sequence

from caqtus.types.expression import Expression
from caqtus.types.parameter import is_quantity, magnitude_in_unit
from ..variable_namespace import VariableNamespace


def evaluate_step_durations(
    steps: Sequence[tuple[str, Expression]], variables: VariableNamespace
) -> list[float]:
    """Returns the durations of each step in seconds.

    Args:
        steps: A sequence of (name, duration) pairs.
        variables: The variables to use for evaluating duration expressions.

    Returns:
        A list of step durations in seconds.
    """

    result = []

    all_variables = variables.dict()

    for name, duration in steps:
        try:
            evaluated = duration.evaluate(all_variables)
        except Exception as e:
            raise ValueError(
                f"Couldn't evaluate duration <{duration}> of step <{name}>"
            ) from e

        if not is_quantity(evaluated):
            raise TypeError(
                f"Duration <{duration}> of step <{name}> is not a quantity "
                f"({evaluated})"
            )

        try:
            seconds = magnitude_in_unit(evaluated, "s")
        except Exception as error:
            raise ValueError(
                f"Duration <{duration}> of step <{name}> can't be converted to seconds "
                f"({evaluated})"
            ) from error
        if seconds < 0:
            raise ValueError(
                f"Duration <{duration}> of step <{name}> is negative ({seconds})"
            )
        result.append(seconds)
    return result



from typing import Optional, Any, assert_never

from PySide6.QtCore import QObject, QModelIndex, Qt
from PySide6.QtGui import QPalette

from caqtus.types.timelane import CameraTimeLane, TakePicture
from caqtus.types.data import DataLabel
from caqtus.types.image import ImageLabel
from .model import TimeLaneModel
from .._icons import get_icon


class CameraTimeLaneModel(TimeLaneModel[CameraTimeLane, None]):
    def __init__(self, name: str, parent: Optional[QObject] = None):
        lane = CameraTimeLane([None])
        super().__init__(name, lane, parent)
        self._brush = None
        palette = QPalette()
        color = palette.text().color()
        self._icon = get_icon("camera", color=color)

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        value = self._lane[index.row()]
        if role == Qt.ItemDataRole.DisplayRole:
            if isinstance(value, TakePicture):
                return value.picture_name
            elif value is None:
                return None
            else:
                assert_never(value)
        elif role == Qt.ItemDataRole.EditRole:
            if isinstance(value, TakePicture):
                return value.picture_name
            elif value is None:
                return ""
            else:
                assert_never(value)
        elif role == Qt.ItemDataRole.ForegroundRole:
            return self._brush
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter
        elif role == Qt.ItemDataRole.DecorationRole:
            if isinstance(value, TakePicture):
                return self._icon
        else:
            return None

    def setData(
        self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole
    ):
        if not index.isValid():
            return False
        if role == Qt.ItemDataRole.EditRole:
            start, stop = self._lane.get_bounds(index.row())
            if isinstance(value, str):
                if value == "":
                    self._lane[start:stop] = None
                elif isinstance(value, str):
                    self._lane[start:stop] = TakePicture(ImageLabel(DataLabel(value)))
                else:
                    raise TypeError(f"Invalid type for value: {type(value)}")
                self.dataChanged.emit(index, index)
                return True
            else:
                raise TypeError(f"Invalid type for value: {type(value)}")
        return False

    def insertRow(self, row, parent: QModelIndex = QModelIndex()) -> bool:
        return self.insert_value(row, None)



from collections.abc import Mapping, Callable

import qtawesome
from PySide6.QtWidgets import QApplication

from caqtus.session import ExperimentSessionMaker
from caqtus.utils.serialization import JSON
from .single_shot_widget import (
    SnapShotMainWindow,
    ViewCreator,
    ShotView,
    SnapShotWindowHandler,
)
from ..._common.sequence_hierarchy import AsyncPathHierarchyView
from ...qtutil import qt_trio


class SnapShot:
    def __init__(
        self,
        session_maker: ExperimentSessionMaker,
        view_creators: Mapping[str, ViewCreator],
        view_dumper: Callable[[ShotView], JSON],
        view_loader: Callable[[JSON], ShotView],
    ):
        app = QApplication.instance()
        self.session_maker = session_maker
        if app is None:
            self.app = QApplication([])
            self.app.setOrganizationName("Caqtus")
            self.app.setApplicationName("Shot Viewer")
            self.app.setWindowIcon(
                qtawesome.icon("mdi6.microscope", size=64, color="grey")
            )
            self.app.setStyle("Fusion")
        else:
            self.app = app

        hierarchy_view = AsyncPathHierarchyView(session_maker)

        self.window = SnapShotMainWindow(
            hierarchy_view=hierarchy_view,
            view_creators=view_creators,
            view_dumper=view_dumper,
            view_loader=view_loader,
        )

    def run(self) -> None:
        self.window.show()

        async def handle():
            handler = SnapShotWindowHandler(self.window, self.session_maker)
            await handler.exec_async()

        qt_trio.run(handle)



import functools
from collections.abc import Callable
from typing import TypeVar

from caqtus.device import DeviceConfiguration
from ._protocol import CondetrolDeviceExtensionProtocol
from .._device_configuration_editor import (
    DeviceConfigurationEditor,
    FormDeviceConfigurationEditor,
)

C = TypeVar("C", bound=DeviceConfiguration)


class CondetrolDeviceExtension(CondetrolDeviceExtensionProtocol):
    def __init__(self):
        self.get_device_configuration_editor = functools.singledispatch(
            get_default_device_configuration_editor
        )

        self.configuration_factories: dict[str, Callable[[], DeviceConfiguration]] = {}

    def register_device_configuration_editor(
        self,
        configuration_type: type[C],
        editor_type: Callable[[C], DeviceConfigurationEditor[C]],
    ) -> None:
        self.get_device_configuration_editor.register(configuration_type)(editor_type)

    def register_configuration_factory(
        self, configuration_label: str, factory: Callable[[], DeviceConfiguration]
    ) -> None:
        self.configuration_factories[configuration_label] = factory

    def available_new_configurations(self) -> set[str]:
        return set(self.configuration_factories.keys())

    def create_new_device_configuration(
        self, configuration_label: str
    ) -> DeviceConfiguration:
        return self.configuration_factories[configuration_label]()


def get_default_device_configuration_editor(
    configuration,
) -> DeviceConfigurationEditor[DeviceConfiguration]:
    if not isinstance(configuration, DeviceConfiguration):
        raise TypeError(f"Expected a DeviceConfiguration, got {type(configuration)}.")
    return FormDeviceConfigurationEditor(configuration)
